@@ -26,6 +26,7 @@
 #include <math.h>
 #include <errno.h>
 #include <assert.h>
+#include <glib.h>
 
 #include "display.h"
 
@@ -114,7 +115,7 @@ localize_expression(char *dest, const char *src, int dest_length, int *cursor)
 
     /* Scan expression looking for numbers and inserting separators */
     output = g_string_sized_new(dest_length);
-    for (c = src, read_cursor = 1; *c; c++, read_cursor++) {
+    for (c = src, read_cursor = 1; *c; c = g_utf8_next_char(c), read_cursor++) {
         /* Insert separators between digits */
         if (*c >= '0' && *c <= '9') {
             /* Read ahead to find the number of digits */
@@ -125,7 +126,7 @@ localize_expression(char *dest, const char *src, int dest_length, int *cursor)
                 }
             }
             
-            g_string_append_c(output, *c);
+            g_string_append_unichar(output, g_utf8_get_char(c));
             
             /* Insert separator after nth digit */
             if (v->display.show_tsep && v->base == DEC &&
@@ -149,7 +150,7 @@ localize_expression(char *dest, const char *src, int dest_length, int *cursor)
         else {
             digit_count = -1;
             after_radix = FALSE;
-            g_string_append_c(output, *c);
+            g_string_append_unichar(output, g_utf8_get_char(c));
         }
     }
     
@@ -217,7 +218,7 @@ gboolean display_get_unsigned_integer(GCDisplay *display, guint64 *value)
     }
     
     /* strtoull() treats the string like a 2's complement number which is not what we want */
-    if(text[0] == '-')
+    if(strncmp(text, "-", strlen("-")) == 0 || strncmp(text, "−", strlen("−")) == 0)
         return FALSE;
 
     *value = g_ascii_strtoull(text, &endptr, bases[v->base]);
@@ -257,13 +258,12 @@ display_set_answer(GCDisplay *display)
 
 
 static void
-display_refresh(GCDisplay *display)
+display_make_text(GCDisplay *display, char *localized, int length, int *cursor)
 {
     int i;
     MPNumber MP_reg;
-    char localized[MAX_LOCALIZED], temp[MAX_LOCALIZED], *str, reg[3];
+    char temp[MAX_LOCALIZED], *str, reg[3];
     GCDisplayState *e;
-    int cursor = display_get_cursor(display);
 
     e = get_state(display);
     if (display_is_empty(display)) {
@@ -286,17 +286,31 @@ display_refresh(GCDisplay *display)
         str = str_replace(str, reg, temp);
     }
 
-    localize_expression(localized, str, MAX_LOCALIZED, &cursor);
-    ui_set_display(localized, cursor);
+    localize_expression(localized, str, length, cursor);
     free(str);
 }
 
 
+static void
+display_refresh(GCDisplay *display)
+{
+    char localized[MAX_LOCALIZED];
+    int cursor;
+    
+    cursor = display_get_cursor(display);
+    display_make_text(display, localized, MAX_LOCALIZED, &cursor);
+    ui_set_display(localized, cursor);
+}
+
+
 void
 display_set_string(GCDisplay *display, const char *value, int cursor)
 {
     GCDisplayState *e;
     
+    if (value[0] == '\0')
+        cursor = -1;
+    
     e = get_state(display);
     free(e->expression);
     e->expression = strdup(value);
@@ -305,6 +319,7 @@ display_set_string(GCDisplay *display, const char *value, int cursor)
     display_refresh(display);
 }
 
+
 void
 display_set_cursor(GCDisplay *display, int cursor)
 {
@@ -315,6 +330,7 @@ display_set_cursor(GCDisplay *display, int cursor)
     display_refresh(display);
 }
 
+
 void
 display_set_error(GCDisplay *display, const char *message)
 {
@@ -427,54 +443,89 @@ display_is_undo_step(GCDisplay *display)
     return(display->h.current != display->h.begin);
 }
 
+
 void
-display_insert(GCDisplay *display, int cursor, const char *text)
+display_insert(GCDisplay *display, int cursor_start, int cursor_end, const char *text)
 {
-    char buf[MAX_DISPLAY], *currentText;
+    char buf[MAX_DISPLAY];
     
-    if (cursor < 0) {
+    if (cursor_start < 0) {
         SNPRINTF(buf, MAX_DISPLAY, "%s%s", display_get_text(display), text);
+        display_set_string(display, buf, -1);
     } else {
-        currentText = ui_get_display();
-        SNPRINTF(buf, MAX_DISPLAY, "%.*s%s%s", cursor, currentText, text, currentText + cursor);
-        cursor += strlen(text);
+        GString *new_text;
+        const char *c;
+        gint cursor, new_cursor;
+        
+        /* Get display text and strip out thousand separators */
+        new_text = g_string_new("");
+        new_cursor = 0;
+        if (cursor_start == 0) {
+            g_string_append(new_text, text);
+            new_cursor += g_utf8_strlen(text, -1);
+        }
+        
+        cursor = 0;
+        for (c = ui_get_display(); *c; c = g_utf8_next_char(c), cursor++) {
+            gboolean use = TRUE;
+            
+            /* Ignore selected part */
+            if (cursor_start != cursor_end && cursor >= cursor_start && cursor < cursor_end)
+                use = FALSE;
+            
+            /* Ignore thousands separators */
+            if (strncmp(c, v->tsep, strlen(v->tsep)) == 0)
+                use = FALSE;
+            
+            /* Copy existing text */
+            if (use) {
+                g_string_append_unichar(new_text, g_utf8_get_char(c));
+                if (cursor < cursor_start)
+                    new_cursor++;
+            }
+            
+            /* Insert text */
+            if ((cursor + 1) == cursor_start) {
+                g_string_append(new_text, text);
+                new_cursor += g_utf8_strlen(text, -1);
+            }
+        }
+        display_set_string(display, new_text->str, new_cursor);
+        g_string_free(new_text, TRUE);
     }
-    display_set_string(display, buf, cursor);
+
 }
 
-void
-display_insert_at_cursor(GCDisplay *display, const char *text)
-{
-    display_insert(display, display_get_cursor(display), text);
-}
 
 void
-display_insert_number(GCDisplay *display, int cursor, const MPNumber *value)
+display_insert_number(GCDisplay *display, int cursor_start, int cursor_end, const MPNumber *value)
 {
     char text[MAX_DISPLAY];
     display_make_number(display, text, MAX_DISPLAY, value, v->base, FALSE);
-    display_insert(display, cursor, text);
-}
-
-void
-display_insert_number_at_cursor(GCDisplay *display, const MPNumber *value)
-{
-    display_insert_number(display, display_get_cursor(display), value);
+    display_insert(display, cursor_start, cursor_end, text);
 }
 
 
 void
-display_backspace(GCDisplay *display)
+display_backspace(GCDisplay *display, int cursor_start, int cursor_end)
 {
-    char buf[MAX_DISPLAY] = "", buf2[MAX_DISPLAY], *t;
+    char buf[MAX_DISPLAY] = "", buf2[MAX_DISPLAY];
     GCDisplayState *e = get_state(display);
     int i, cursor;
     MPNumber MP_reg;
     
-    cursor = display_get_cursor(display);
+    /* Can't delete empty display */
+    if (display_is_empty(display))
+        return;
 
+    cursor = display_get_cursor(display);
+    
     /* If cursor is at end of the line then delete the last character preserving accuracy */
-    if (cursor < 0) {
+    if (cursor_start < 0) {
+        int len;
+        
+        len = g_utf8_strlen(ui_get_display(), -1);
+        
         if (exp_has_postfix(e->expression, "Ans")) {
             display_make_number(display, buf, MAX_DISPLAY, &e->ans, v->base, FALSE);
             e->expression = str_replace(e->expression, "Ans", buf);
@@ -484,38 +535,32 @@ display_backspace(GCDisplay *display)
                 if (exp_has_postfix(e->expression, buf)) {
                     register_get(i, &MP_reg);
                     display_make_number(display, buf2, MAX_DISPLAY, &MP_reg, v->base, FALSE);
-                    /* Remove "Rx" postfix and replace with backspaced number */
-                    SNPRINTF(buf, MAX_DISPLAY, "%.*s%s", strlen(e->expression) - 2, e->expression - 3, buf2);
-                    display_set_string(display, buf, cursor - 1);
-                    return;
+                    SNPRINTF(buf, MAX_DISPLAY, "%.*s%s", strlen(e->expression) - 2, e->expression - 2, buf2);
+                    break;
                 }
             }
         }
 
-        SNPRINTF(buf, MAX_DISPLAY, "%.*s", strlen(e->expression) - 1, e->expression);
-    } else if (cursor > 0) {
-        t = ui_get_display();
-        SNPRINTF(buf, MAX_DISPLAY, "%.*s%s", cursor - 1, t, t + cursor);
-    } else {
-        return; /* At the start of the line */
+        display_insert(display, len - 1, len, "");
+    } else if (cursor_start != cursor_end) {
+        display_insert(display, cursor_start, cursor_end, "");
+    } else if (cursor_start > 0) {
+        display_insert(display, cursor_start - 1, cursor_start, "");
     }
 
-    display_set_string(display, buf, cursor - 1);
 }
 
 void
-display_delete(GCDisplay *display)
+display_delete(GCDisplay *display, int cursor_start, int cursor_end)
 {
-    char buf[MAX_DISPLAY] = "", *text;
-    int cursor = display_get_cursor(display);    
-    
-    if (cursor >= 0) {
-        text = ui_get_display();
-        SNPRINTF(buf, MAX_DISPLAY, "%.*s%s", cursor, text, text + cursor + 1);
-        display_set_string(display, buf, cursor);
-    }
+    /* Delete selected block */
+    if (cursor_start != cursor_end)
+        display_insert(display, cursor_start, cursor_end, "");
+    else if (cursor_start >= 0)
+        display_insert(display, cursor_start, cursor_start + 1, "");
 }
 
+
 void
 display_surround(GCDisplay *display, const char *prefix, const char *suffix)
 {
@@ -532,6 +577,7 @@ display_is_empty(GCDisplay *display)
     return strcmp(display_get_text(display), "") == 0;
 }
 
+
 gboolean
 display_is_result(GCDisplay *display)
 {
@@ -541,6 +587,7 @@ display_is_result(GCDisplay *display)
     return FALSE;
 }
 
+
 gboolean
 display_is_usable_number(GCDisplay *display, MPNumber *MPnum)
 {
@@ -658,7 +705,8 @@ make_eng_sci(GCDisplay *display, char *target, int target_len, const MPNumber *M
     mp_abs(MPnumber, &MPval);
     mp_set_from_integer(0, &MP1);
     if (mp_is_less_than(MPnumber, &MP1)) {
-        *optr++ = '-';
+        strcpy(optr, "−");
+        optr += strlen("−");
     }
     mp_set_from_mp(&MPval, &MPmant);
 
@@ -705,7 +753,8 @@ make_eng_sci(GCDisplay *display, char *target, int target_len, const MPNumber *M
  
     if (exp < 0) {
         exp = -exp;
-        *optr++ = '-';
+        strcpy(optr, "−");
+        optr += strlen("−");
     } else {
         *optr++ = '+';
     }
